# -*- coding: utf-8 -*-
# vim: set et sts=2:
##############################################################################
#
# Copyright (c) 2018 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import os
import sys
import logging
import time
import signal
import importlib
import re
import traceback
import psutil
import subprocess
import functools
import slapos.slap
from abc import ABCMeta, abstractmethod
from multiprocessing import Process, Queue as MQueue
import Queue
from datetime import datetime, timedelta
from slapos.grid.promise import interface
from slapos.grid.utils import dropPrivileges
from zope import interface as zope_interface

PROMISE_STATE_FOLDER_NAME = '.promises'

class PromiseError(Exception):
  pass

class BaseResult(object):
  def __init__(self, problem=False, message=None, date=None):
    self.__problem = problem
    self.__message = message
    self.__date = date
    if self.__date is None:
      self.__date = datetime.utcnow()

  def hasFailed(self):
    return self.__problem

  @property
  def type(self):
    return "Base Result"

  @property
  def message(self):
    return self.__message

  @property
  def date(self):
    return self.__date

class TestResult(BaseResult):

  @property
  def type(self):
    return "Test Result"

class AnomalyResult(BaseResult):

  @property
  def type(self):
    return "Anomaly Result"

class PromiseQueueResult(object):

  def __init__(self, path, name, title, item, execution_time=0):
    self.path = path
    self.name = name
    self.item = item
    self.title = title
    self.execution_time = execution_time

class GenericPromise(object):

  # Abstract class
  __metaclass__ = ABCMeta

  def __init__(self, config):
    self.__config = config

    self.__log_folder = self.__config.pop('log-folder', None)
    self.__partition_folder = self.__config.pop('partition-folder', None)
    self.__periodicity = self.__config.pop('periodicity', 2)
    self.__debug = self.__config.pop('debug', True)
    self.__name = self.__config.pop('name', None)
    self.__promise_path = self.__config.pop('path', None)
    self.__queue = self.__config.pop('queue', None)
    self.__logger_buffer = None

    self._validateConf()
    self._configureLogger()
    for key, value in config.items():
      setattr(self, key.replace('-', '_'), value)

  def _configureLogger(self):
    self.logger = logging.getLogger(self.__name)
    for handler in self.logger.handlers:
      self.logger.removeHandler(handler)
    if self.__log_folder is None:
      # configure logger with StringIO
      import cStringIO
      self.__logger_buffer = cStringIO.StringIO()
      logger_handler = logging.StreamHandler(self.__logger_buffer)
      self.__log_file = None
    else:
      self.__log_file = os.path.join(
        self.__log_folder,
        '%s.log' % self.__title
      )
      logger_handler = logging.FileHandler(self.__log_file)

    self.logger.setLevel(logging.DEBUG if self.__debug else logging.INFO)
    logger_handler.setFormatter(
      fmt=logging.Formatter("%(asctime)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S")
    )
    self.logger.addHandler(logger_handler)

  def _validateConf(self):
    if self.__queue is None:
      raise ValueError("Queue object is not set in configuration")
    if self.__name is None:
      raise ValueError("Monitor name is not set in configuration")
    self.__title = os.path.splitext(self.__name)[0]
    if self.__promise_path is None:
      raise ValueError("Promise path is not set in configuration")
    if self.__partition_folder is None:
      raise ValueError("Monitor partition folder is not set in configuration")

  def getConfig(self, key, default=None):
    return self.__config.get(key, default)

  def setConfig(self, key, value):
    self.__config[key] = value

  def getTitle(self):
    return self.__title

  def getName(self):
    return self.__name

  def getLogFile(self):
    return self.__log_file

  def getLogFolder(self):
    return self.__log_folder

  def getPartitionFolder(self):
    return self.__partition_folder

  def getPromiseFile(self):
    return self.__promise_path

  def setPeriodicity(self, minute):
    if minute <= 0:
      raise ValueError("Cannot set promise periodicity to a value less than 1")
    self.__periodicity = minute

  def getPeriodicity(self):
    return self.__periodicity

  def __bang(self, message):
    """
      Call bang if requested
    """
    if self.__config.has_key('master_url') and \
        self.__config.has_key('partition-id') and \
        self.__config.has_key('computer-id'):

      slap = slapos.slap.slap()
      slap.initializeConnection(
          self.__config['master-url'],
          self.__config.get('partition-key'),
          self.__config.get('partition-cert'),
      )
      computer_partition = slap.registerComputerPartition(
          self.__config['computer-id'],
          self.__config['partition-id'],
      )
      computer_partition.bang(message)

  def __getLogRegex(self):
    return re.compile(r"(\d{4}-\d{2}-\d{2}\s+\d{2}:\d{2}:\d{2})\s+\-?\s*(\w{4,7})\s+\-?\s*(.*)")

  def __getResultFromString(self, result_string, only_failure=False):
    line_list = result_string.split('\n')
    result_list = []
    line_part = ""
    regex = self.__getLogRegex()
    for line in line_list:
      if not line:
        continue
      match = regex.match(line)
      if match is not None:
        if not only_failure or (only_failure and match.groups()[1] == 'ERROR'):
          result_list.append({
            'date': datetime.strptime(match.groups()[0], '%Y-%m-%d %H:%M:%S'),
            'status': match.groups()[1],
            'message': (match.groups()[2] + line_part).strip(),
          })
        line_part = ""
      else:
        line_part += '\n' + line
    result_list.reverse()

    return result_list

  def __checkLogDateInNewPromise(self, log_date, previous_date):
    if previous_date is None:
      return True

    diff = previous_date - log_date
    if diff.seconds >= self.getConfig('promise-timeout'):
      # The promise run in max 20 seconds so this was not in the same process
      return True
    return False

  def getLastPromiseResultList(self, latest_minute=0, result_count=1,
      only_failure=False):
    """
      Return the latest log result of the promise starting from the most recent

      @param last_minute: the number of minutes in the past. If last_minute is
        1, it will return the log of the latest minute execution. 0 => disabled
      @param only_failure: only return the lines which contain failures.
      @param result_count: maximum number of promise result to check, will not
        exceed the `latest_minute`
      @return Return a dict of logs. The format is
        [{"date": "DATE", "status": "STATUS", "message": MESSAGE}, ]
    """

    if self.__log_file is None:
      if self.__logger_buffer is not None:
        return self.__getResultFromString(self.__logger_buffer.getvalue(),
                                          only_failure)
      else:
        return []

    if not os.path.exists(self.__log_file):
      return []

    regex = self.__getLogRegex()
    max_date_string = ""
    if latest_minute > 0:
      date = datetime.now() - timedelta(minutes=latest_minute)
      max_date_string = date.strftime('%Y-%m-%d %H:%M:%S')

    line_list = []
    previous_date = None
    promise_count = 0
    with open(self.__log_file, 'r') as f:
      offset = 0
      f.seek(0, 2)
      size = f.tell() * -1
      line = line_part = ""
      while offset > size:
        offset -= 1
        f.seek(offset, 2)
        char = f.read(1)
        if char != '\n':
          line = char + line
        if char == '\n' or offset == size:
          # Add new line
          if offset == -1:
            continue
          if line != "":
            result = regex.match(line)
            if result is not None:
              if max_date_string and result.groups()[0] <= max_date_string:
                break
              log_date = datetime.strptime(result.groups()[0], '%Y-%m-%d %H:%M:%S')
              if self.__checkLogDateInNewPromise(log_date, previous_date):
                promise_count += 1
                previous_date = log_date
              if promise_count > result_count:
                break
              if not only_failure or \
                  (only_failure and result.groups()[1] == 'ERROR'):
                line_list.append({
                  'date': log_date,
                  'status': result.groups()[1],
                  'message': (result.groups()[2] + line_part).strip(),
                })
            else:
              line_part = '\n' + line + line_part
              line = ""
              continue
            line = line_part = ""
    return line_list

  def defaultTest(self, result_count=1, failure_amount=1, latest_minute=0,
      is_anomaly=False):
    """
      Test if the latest messages contain `failure_amount` failures.

      @param result_count: maximum number of promise result to check, will not
        exceed the `latest_minute`
      @param latest_minute: test the result from now to the latest X minutes in
        the past.
      @param failure_amount: fail is this amount of failure is found in result
      @param is_anomaly: Say if the result is an AnomalyResult of TestResult
    """

    problem = True
    message = ""
    module = TestResult if not is_anomaly else AnomalyResult
    latest_result_list = self.getLastPromiseResultList(
      result_count=result_count,
      latest_minute=latest_minute,
      only_failure=False
    )
    result_size = len(latest_result_list)
    if result_size == 0:
      return module(problem=False, message="No result found!")
    if latest_result_list[0]['status'] != 'ERROR':
      # latest execution is OK
      return module(problem=False, message=latest_result_list[0]['message'])

    i = 1
    failure_found = 1
    while i < result_size and failure_found < failure_amount:
      if latest_result_list[i]['status'] == 'ERROR':
        failure_found += 1
      i += 1

    if failure_found != failure_amount:
      return module(problem=False, message=latest_result_list[0]['message'])
    return module(problem=problem, message=latest_result_list[0]['message'])

  @abstractmethod
  def sense(self):
    """Run the promise code and log the result"""

  def anomaly(self):
    """Called to detect if there is an anomaly which require to bang."""
    raise NotImplementedError("'anomaly' method is not implemented yet!")

  def test(self):
    """Test promise and say if problem is detected or not"""
    raise NotImplementedError("'test' method is not implemented yet!")

  def run(self, can_bang=True):
    """
      Method called to run the Promise
      @param can_bang: Set to True if bang can be called, this parameter should
        be set to False if bang is already called by another promise.
    """
    try:
      self.sense()
    except Exception, e:
      # log the result
      self.logger.error(str(e))
    if self.getConfig('check-anomaly', False):
      # run sense, anomaly
      try:
        result = self.anomaly()
        if result is None:
          raise ValueError("Promise anomaly method returned 'None'")
      except Exception, e:
        result = AnomalyResult(problem=True, message=str(e))
      else:
        if result.hasFailed() and can_bang:
          self.__bang("Promise %s is failing" % self.__title)
    else:
      # run sense, test
      try:
        result = self.test()
        if result is None:
          raise ValueError("Promise test method returned 'None'")
      except Exception, e:
        result = TestResult(problem=True, message=str(e))

    if self.__logger_buffer is not None:
      self.__logger_buffer.close()

    # send the result of this promise
    self.__queue.put(PromiseQueueResult(
        path=self.__promise_path,
        name=self.__name,
        title=self.__title,
        item=result
      ), True)


class PromiseWrapper(GenericPromise):
  """
    A wrapper promise used to run old promises style and bash promises
  """

  zope_interface.implements(interface.IPromise)

  def __init__(self, config):
    GenericPromise.__init__(self, config)
    self.setConfig("check-anomaly", False)
    self.setPeriodicity(minute=2)

  @staticmethod
  def terminate(name, logger, process, signum, frame):
    if signum in [signal.SIGINT, signal.SIGTERM] and process:
      logger.info("Terminating promise process %r" % name)
      try:
        # make sure we kill the process on timeout
        process.terminate()
      except Exception, e:
        logger.error(traceback.format_exc())

  def sense(self):
    promise_process = subprocess.Popen(
        [self.getPromiseFile()],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=self.getPartitionFolder()
    )
    handler = functools.partial(self.terminate, self.getName(), self.logger,
                                promise_process)
    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)

    output, error = promise_process.communicate()
    message = output or ""
    if error:
      message += "\n" + error
    if promise_process.returncode != 0:
      self.logger.error(message.strip())
    else:
      self.logger.info(message.strip())

  def test(self):
    return self.defaultTest(result_count=1, failure_amount=1, is_anomaly=False)


class PromiseRunner(Process):

  """
    Run a promise in a new Process
  """

  def __init__(self, promise_instance, logger, allow_bang=True, uid=None,
      gid=None, cwd=None):
    """
      Initialise Promise Runner

      @param promise_instance: Promise instance from GenericPromise class
      @param allow_bang: Bolean saying if bang should be called in case of
        anomaly failure.
    """
    Process.__init__(self)
    self.promise = promise_instance
    self.logger = logger
    self.allow_bang = allow_bang
    self.uid = uid
    self.gid = gid
    self.cwd = cwd
    # set this to True, so children process will be killed when exit
    self.daemon = True

  def run(self):
    if self.uid and self.gid:
      dropPrivileges(self.uid, self.gid, logger=self.logger)
    if self.cwd is not None:
      os.chdir(self.cwd)
    self.promise.run(self.allow_bang)


class PromiseLauncher(object):

  def __init__(self, config=None, logger=None, save_method=None,
      extra_config=None):
    """
      Promise launcher will run promises

      @param config_file: A file containing configurations
      @param save_method: A method to call for saving promise result. If
        None, the promise excution will raise in case of failure.
      @param logger: Set the logger to use, if None a logger will be configured
        to console.
      @param config: A configuration dict to use. Values send here will
        overwrite configs from `config_file`. Expected values in config are:
        promise-timeout
          Maximum promise execution time before timeout. Default: 20
        partition-folder
          Base path of the partition
        promise-folder
          Promises folder, all promises scripts will be imported from that folder
        old-promise-folder
          Old promises folder, where to find bash, shell and standard promises
        log-folder
          Folder where promises will write logs. Can be None
        check-anomaly
          Ask to check anomaly instead of test. Default: False
        debug
          Configure loggin in debug mode. Default: True
        master-url
          SlapOS Master service URL
        partition-cert
          Computer Partition Certificate file
        partition-key
          Computer Partition key file
        partition-id
          Computer Partition ID, ex: slappart13
        computer-id
          Computer ID, ex: COMP-1234
        uid
          User UID
        gid
          User GID
        debug
          If True, show Promise consumption and execution time information, etc
        run-only-promise-list
          A list of promise from plugins directory that will be executed
        slapgrid-mode
          Set to True if promise launcher is executed by slapgrid.
        force
          Set to True if force run promises without check their periodicity
      @param extra_config: A dictionary of extra config to send to all promises
    """

    self.save_method = save_method
    self.__config = {
      'promise-timeout': 20,
      'promise-folder': None,
      'old-promise-folder': None,
      'log-folder': None,
      'partition-folder': None,
      'debug': False,
      'uid': None,
      'gid': None,
      'master-url': None,
      'partition-cert': None,
      'partition-key': None,
      'partition-id': None,
      'computer-id': None,
      'check-anomaly': False,
      'force': False,
      'run-only-promise-list': None,
      'slapgrid-mode': False
    }
    if config is not None:
      self.__config.update(config)
    self.__extra_config = {}
    if extra_config is not None:
      self.__extra_config = extra_config

    for key, value in self.__config.items():
      setattr(self, key.replace('-', '_'), value or None)

    if self.promise_folder is None:
      raise ValueError("Promise folder is missing in configuration!")
    if self.partition_folder is None:
      raise ValueError("Partition folder is missing in configuration!")

    if logger is None:
      self.logger = logging.getLogger(__name__)
      self.logger.setLevel(logging.DEBUG if self.debug else logging.INFO)
      if len(self.logger.handlers) == 0 or \
          not isinstance(self.logger.handlers[0], logging.StreamHandler):
        handler = logging.StreamHandler()
        handler.setFormatter(
          logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        )
        self.logger.addHandler(handler)
    else:
      self.logger = logger

    self.queue_result = MQueue()
    self.bang_called = False
    self.run_status_folder = os.path.join(
      self.partition_folder,
      PROMISE_STATE_FOLDER_NAME
    )
    if not os.path.exists(self.run_status_folder):
      os.mkdir(self.run_status_folder)
    if self.slapgrid_mode:
      self.force = True

  def _isPromisePeriodicityMatch(self, promise_name, periodicity):
    """
      Return True if promise should be run now, considering given the execution
        periodicity in minutes
    """
    state_file = os.path.join(self.run_status_folder, promise_name)
    if os.path.exists(state_file) and os.stat(state_file).st_size:
      with open(state_file) as f:
        try:
          latest_timestamp = float(f.read())
          current_timediff = (time.time() - latest_timestamp) / 60.0
          if current_timediff >= periodicity:
            return True
          self.logger.debug("Skip Promise %r. periodicity=%s, time_diff=%s" % (
              promise_name, periodicity, current_timediff))
        except ValueError:
          # if the file is broken, run the promise and regenerated it
          return True
        else:
          return False
    return True

  def _setPromiseLatestRunTime(self, promise_name):
    """
      Save the promise execution timestamp
    """
    state_file = os.path.join(self.run_status_folder, promise_name)
    with open(state_file, 'w') as f:
      f.write(str(time.time()))

  def _loadPromiseModule(self, promise_name):
    """Load a promise from promises directory."""

    if re.match(r'[a-zA-Z_]', promise_name) is None:
      self.logger.error("Promise plugin name %r is not valid" % promise_name)

    promise_module = importlib.import_module(os.path.splitext(promise_name)[0])
    if not hasattr(promise_module, "RunPromise"):
      raise AttributeError("Class RunPromise not found in promise" \
        "%s" % promise_name)
    if not interface.IPromise.implementedBy(promise_module.RunPromise):
      raise RuntimeError("RunPromise class in %s must implements 'IPromise'" \
        " interface. zope_interface.implements(interface.IPromise) is" \
        " missing ?" % promise_name)
    if not issubclass(promise_module.RunPromise, GenericPromise):
      raise RuntimeError("RunPromise class is not a subclass of" \
        "GenericPromise class.")

    return promise_module

  def _getErrorPromiseResult(self, promise_instance, promise_name, message,
      execution_time=0):
    if self.check_anomaly:
      result = AnomalyResult(problem=True, message=message)
    else:
      result = TestResult(problem=True, message=message)
    return PromiseQueueResult(
      item=result,
      path=os.path.join(self.promise_folder, promise_name),
      name=promise_name,
      title=promise_instance.getTitle(),
      execution_time=execution_time
    )

  def _launchPromise(self, promise_name, argument_dict, promise_module=None):
    """
      Launch the promise and save the result if `self.save_method` is not None
        If no save method is set, raise PromiseError in case of failure
    """
    try:
      if promise_module is None:
        promise_instance = PromiseWrapper(argument_dict)
      else:
        promise_instance = promise_module.RunPromise(argument_dict)
    except Exception:
      # only print traceback to not prevent run other promises
      self.logger.error(traceback.format_exc())
      self.logger.warning("Promise %s skipped." % promise_name)
      return

    if not self.force and not self._isPromisePeriodicityMatch(promise_name,
        promise_instance.getPeriodicity()):
      return
    self.logger.info("Checking promise %s..." % promise_name)

    promise_process = PromiseRunner(
      promise_instance,
      logger=self.logger,
      allow_bang=not self.bang_called and self.check_anomaly,
      uid=self.uid,
      gid=self.gid,
      cwd=self.partition_folder
    )

    promise_process.start()
    if not self.slapgrid_mode:
      self._setPromiseLatestRunTime(promise_name)
    sleep_time = 0.1
    increment_limit = int(self.promise_timeout / sleep_time)
    execution_time = self.promise_timeout
    ps_profile = False
    if self.debug:
      try:
        psutil_process = psutil.Process(promise_process.pid)
        ps_profile = True
      except psutil.NoSuchProcess:
        # process is gone
        pass
    for current_increment in range(0, increment_limit):
      execution_time = (current_increment + 1) * sleep_time
      if not promise_process.is_alive():
        try:
          queue_item = self.queue_result.get(True, 1)
        except Queue.Empty:
          # no result found in process result Queue
          if self.save_method is None:
            raise PromiseError("No output returned by the promise")
          queue_item = self._getErrorPromiseResult(
            promise_instance,
            promise_name=promise_name,
            messsage="No output returned by the promise",
            execution_time=execution_time
          )
          self.save_method(result=queue_item)
          break

        if queue_item.item.hasFailed():
          if self.save_method is None:
            raise PromiseError(queue_item.item.message)
          elif isinstance(queue_item.item, AnomalyResult):
            # stop to bang as it was called
            self.bang_called = True

        if self.save_method is not None:
          queue_item.execution_time = execution_time
          self.save_method(result=queue_item)
        break

      if ps_profile:
        try:
          io_counter = psutil_process.io_counters()
          self.logger.debug(
            "[t=%ss] CPU: %s%%, MEM: %s MB (%s%%), DISK: %s Read - %s Write" % (
              execution_time,
              psutil_process.cpu_percent(),
              psutil_process.memory_info().rss / float(2 ** 20),
              round(psutil_process.memory_percent(), 4),
              io_counter.read_count,
              io_counter.write_count
            )
          )
        except (psutil.AccessDenied, psutil.NoSuchProcess):
          # defunct process will raise AccessDenied
          pass
      time.sleep(sleep_time)
    else:
      promise_process.terminate()
      message = 'Promise timed out after %s seconds' % self.promise_timeout
      if self.save_method is None:
        raise PromiseError(message)
      else:
        self.save_method(result=self._getErrorPromiseResult(
          promise_instance,
          promise_name=promise_name,
          message=message,
          execution_time=execution_time
        ))
    if self.debug:
      self.logger.debug("Finished promise %r in %s second(s)." % (
                       promise_name, execution_time))

  def run(self):
    """
      Run all promises
    """
    promise_list = []
    base_config = {
      'log-folder': self.log_folder,
      'partition-folder': self.partition_folder,
      'debug': self.debug,
      'promise-timeout': self.promise_timeout,
      'slapgrid-mode': self.slapgrid_mode,
      'check-anomaly': self.check_anomaly,
      'master-url': self.master_url,
      'partition-cert': self.partition_cert,
      'partition-key': self.partition_key,
      'partition-id': self.partition_id,
      'computer-id': self.computer_id,
      'queue': self.queue_result,
    }
    base_config.update(self.__extra_config)

    if os.path.exists(self.promise_folder) and os.path.isdir(self.promise_folder):
      # if there is no __init file, add it
      init_file = os.path.join(self.promise_folder, '__init__.py')
      if not os.path.exists(init_file):
        with open(init_file, 'w') as f:
          f.write("")
        os.chmod(init_file, 0644)
      if sys.path[0] != self.promise_folder:
        sys.path[0:0] = [self.promise_folder]

      promise_list = []
      # load all promises so we can catch import errors before launch them
      for promise_name in os.listdir(self.promise_folder):
        if promise_name.startswith('__init__') or \
            not promise_name.endswith('.py'):
          continue
        if self.run_only_promise_list is not None and not \
            promise_name in self.run_only_promise_list:
          continue
        promise_list.append((promise_name,
                             self._loadPromiseModule(promise_name)))

      for name, module in promise_list:
        promise_path = os.path.join(self.promise_folder, name)
        config = {
          'path': promise_path,
          'name': name
        }
        if module.__file__ != promise_path:
          # cached module need to be updated
          module = reload(module)
        #if os.path.getmtime(promise_path) > os.path.getmtime('%sc' % promise_path):
        #  module = reload(module)

        config.update(base_config)
        self._launchPromise(name, config, module)

    if not self.run_only_promise_list and os.path.exists(self.old_promise_folder) \
        and os.path.isdir(self.old_promise_folder):
      # run old promise styles
      for promise_name in os.listdir(self.old_promise_folder):
        promise_path = os.path.join(self.old_promise_folder, promise_name)
        if not os.path.isfile(promise_path) or \
            not os.access(promise_path, os.X_OK):
          self.logger.warning("Bad promise file at %r." % promise_path)
          continue

        config = {
          'path': promise_path,
          'name': promise_name
        }
        config.update(base_config)
        # We will use promise wrapper to run this
        self._launchPromise(promise_name, config)
